#!/usr/bin/env python3

import glob
import os.path
import pathlib
import platform
import sys
import sysconfig

from cx_Freeze import Executable, setup


def get_non_python_libs():
    """Returns list of tuples containing extra dependencies required to run
    meld on current platform.
    Every pair corresponds to a single library file.
    First tuple item is path in local filesystem during build.
    Second tuple item correspond to path expected in meld installation
    relative to meld prefix.
    Note that for returned dynamic libraries and executables dependencies
    are expected to be resolved by caller, for example by cx_freeze.
    """
    local_bin = os.path.join(sys.prefix, "bin")

    inst_root = []  # local paths of files "to put at freezed root"
    inst_lib = []  # local paths of files "to put at freezed 'lib' subdir"

    if 'mingw' in sysconfig.get_platform():
        # dll imported by dll dependencies expected to be auto-resolved later
        inst_root = [os.path.join(local_bin, 'libgtksourceview-4-0.dll')]

        # required for communicating multiple instances
        inst_lib.append(os.path.join(local_bin, 'gdbus.exe'))

        # gspawn-helper is needed for Gtk.show_uri function
        if platform.architecture()[0] == '32bit':
            inst_lib.append(os.path.join(local_bin, 'gspawn-win32-helper.exe'))
        else:
            inst_lib.append(os.path.join(local_bin, 'gspawn-win64-helper.exe'))

    return [
        (f, os.path.basename(f)) for f in inst_root
    ] + [
        (f, os.path.join('lib', os.path.basename(f))) for f in inst_lib
    ]


gtk_data_dirs = [
    'etc/fonts',
    'etc/gtk-3.0',
    'lib/gdk-pixbuf-2.0',
    'lib/girepository-1.0',
    'share/fontconfig',
    'share/glib-2.0',
    'share/gtksourceview-4',
    'share/icons',
]

gtk_data_files = []
for data_dir in gtk_data_dirs:
    local_data_dir = os.path.join(sys.prefix, data_dir)

    for local_data_subdir, dirs, files in os.walk(local_data_dir):
        data_subdir = os.path.relpath(local_data_subdir, local_data_dir)
        gtk_data_files.append((
            os.path.join(data_dir, data_subdir),
            [os.path.join(local_data_subdir, file) for file in files],
        ))

manually_added_libs = {
    # add libgdk_pixbuf-2.0-0.dll manually to forbid auto-pulling of gdiplus.dll
    "libgdk_pixbuf-2.0-0.dll": os.path.join(sys.prefix, 'bin'),
    # librsvg is needed for SVG loading in gdkpixbuf
    "librsvg-2-2.dll": os.path.join(sys.prefix, 'bin'),
}

for lib, possible_path in manually_added_libs.items():
    local_lib = os.path.join(possible_path, lib)
    if os.path.isfile(local_lib):
        gtk_data_files.append((os.path.dirname(lib), [local_lib]))

build_exe_options = {
    "includes": ["gi"],
    "excludes": ["tkinter"],
    "packages": ["gi", "weakref"],
    "include_files": get_non_python_libs(),
    "bin_excludes": list(manually_added_libs.keys()),
    "zip_exclude_packages": [],
    "zip_include_packages": ["*"],
}


# Create our registry key, and fill with install directory and exe
registry_table = [
    ('MeldKLM', 2, r'SOFTWARE\Meld', '*', None, 'TARGETDIR'),
    ('MeldInstallDir', 2, r'SOFTWARE\Meld', 'InstallDir', '[TARGETDIR]', 'TARGETDIR'),
    ('MeldExecutable', 2, r'SOFTWARE\Meld', 'Executable', '[TARGETDIR]Meld.exe', 'TARGETDIR'),
]

# Provide the locator and app search to give MSI the existing install directory
# for future upgrades
reg_locator_table = [
    ('MeldInstallDirLocate', 2, r'SOFTWARE\Meld', 'InstallDir', 0),
]
app_search_table = [('TARGETDIR', 'MeldInstallDirLocate')]

msi_data = {
    'Registry': registry_table,
    'RegLocator': reg_locator_table,
    'AppSearch': app_search_table,
}

bdist_msi_options = {
    "upgrade_code": "{1d303789-b4e2-4d6e-9515-c301e155cd50}",
    "data": msi_data,
    "all_users": True,
    "add_to_path": True,
    "install_icon": "data/icons/org.gnome.meld.ico",
}

executable_options = {
    "script": "bin/meld",
    "icon": "data/icons/org.gnome.meld.ico",
}
console_executable_options = dict(executable_options)

if 'mingw' in sysconfig.get_platform():
    executable_options.update({
        "base": "Win32GUI",  # comment to build console version to see stderr
        "target_name": "Meld.exe",
        "shortcut_name": "Meld",
        "shortcut_dir": "ProgramMenuFolder",
    })
    console_executable_options.update({
        "target_name": "MeldConsole.exe",
    })

# Copy conf.py in place if necessary
base_path = pathlib.Path(__file__).parent
conf_path = base_path / 'meld' / 'conf.py'

if not conf_path.exists():
    import shutil
    shutil.copyfile(conf_path.with_suffix('.py.in'), conf_path)

import meld.build_helpers  # noqa: E402
import meld.conf  # noqa: E402


setup(
    name="Meld",
    version=meld.conf.__version__,
    description='Visual diff and merge tool',
    author='The Meld project',
    author_email='meld-list@gnome.org',
    maintainer='Kai Willadsen',
    url='http://meldmerge.org',
    license='GPLv2+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Desktop Environment :: Gnome',
        'Topic :: Software Development',
        'Topic :: Software Development :: Version Control',
    ],
    keywords=['diff', 'merge'],
    options={
        "build_exe": build_exe_options,
        "bdist_msi": bdist_msi_options,
        #  cx_freeze + bdist_dumb fails on non-empty prefix
        "install": {"prefix": "."},
        #  freezed binary doesn't use source files, they are only for humans
        "install_lib": {"compile": False},
    },
    executables=[
        Executable(**executable_options),
        Executable(**console_executable_options),
    ],
    packages=[
        'meld',
        'meld.matchers',
        'meld.ui',
        'meld.vc',
    ],
    package_data={
        'meld': ['README', 'COPYING', 'NEWS'],
        'meld.vc': ['README', 'COPYING'],
    },
    scripts=['bin/meld'],
    data_files=[
        ('share/man/man1',
         ['data/meld.1']
         ),
        ('share/doc/meld-' + meld.conf.__version__,
         ['COPYING', 'NEWS']
         ),
        ('share/meld/icons',
         glob.glob("data/icons/*.png") +
         glob.glob("data/icons/COPYING*")
         ),
        ('share/meld/styles',
         glob.glob("data/styles/*.xml")
         ),
        ('share/meld/ui',
         glob.glob("data/ui/*.ui") + glob.glob("data/ui/*.xml")
         ),
    ] + gtk_data_files,
    cmdclass={
        "build_i18n": meld.build_helpers.build_i18n,
        "build_help": meld.build_helpers.build_help,
        "build_icons": meld.build_helpers.build_icons,
        "build_data": meld.build_helpers.build_data,
    }
)




import pytest


@pytest.mark.parametrize("patterns,filename,expected_match", [
    ([r'*.csv'], 'foo.csv', True),
    ([r'*.cvs'], 'foo.csv', False),
    ([r'*.csv *.xml'], 'foo.csv', True),
    ([r'*.csv *.xml'], 'foo.xml', True),
    ([r'*.csv', r'*.xml'], 'foo.csv', True),
    ([r'*.csv', r'*.xml'], 'foo.xml', True),
    ([r'*.csv', r'*.xml'], 'dumbtest', False),
    ([r'thing*csv'], 'thingcsv', True),
    ([r'thing*csv'], 'thingwhatevercsv', True),
    ([r'thing*csv'], 'csvthing', False),
])
def test_file_filters(patterns, filename, expected_match):
    from meld.filters import FilterEntry

    filters = [
        FilterEntry.new_from_gsetting(("name", True, p), FilterEntry.SHELL)
        for p in patterns
    ]

    # All of the dirdiff logic is "does this match any filter", so
    # that's what we test here, even if it looks a bit weird.
    match = any(f.filter.match(filename) for f in filters)
    assert match == expected_match


@pytest.mark.parametrize("pattern", [
    r'*.foo*',  # Trailing wildcard
    r'\xfoo',  # Invalid escape
])
def test_bad_regex_compilation(pattern):
    from meld.filters import FilterEntry

    f = FilterEntry.new_from_gsetting(
        ("name", True, pattern), FilterEntry.REGEX)
    assert f.filter is None



#!/usr/bin/env python3

import sys
from compileall import compile_dir
from os import environ, path
from subprocess import call

if not environ.get('DESTDIR', ''):
    PREFIX = environ.get('MESON_INSTALL_PREFIX', '/usr/local')
    DATA_DIR = path.join(PREFIX, 'share')
    print('Updating icon cache...')
    call(['gtk-update-icon-cache', '-qtf', path.join(DATA_DIR, 'icons', 'hicolor')])
    print("Compiling new schemas")
    call(["glib-compile-schemas", path.join(DATA_DIR, 'glib-2.0', 'schemas')])
    print("Updating desktop database")
    call(["update-desktop-database", path.join(DATA_DIR, 'applications')])

# Byte-compilation is enabled by passing the site-packages path to this script
if len(sys.argv) > 1:
    print('Byte-compiling Python module...')
    python_source_install_path = sys.argv[1]
    compile_dir(path.join(python_source_install_path, 'meld'), optimize=1)



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2011-2016 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


def tree_path_as_tuple(path):
    """Get the path indices as a tuple

    This helper only exists because we often want to use tree paths
    as set members or dictionary keys, and this is a convenient option.
    """
    return tuple(path.get_indices())


def tree_path_prev(path):
    if not path or path[-1] == 0:
        return None
    return path[:-1] + [path[-1] - 1]


def tree_path_up(path):
    if not path:
        return None
    return path[:-1]


def valid_path(model, path):
    try:
        model.get_iter(path)
        return True
    except ValueError:
        return False


def refocus_deleted_path(model, path):
    # Since the passed path has been deleted, either the path is now a
    # valid successor, or there are no successors. If valid, return it.
    # If not, and the path has a predecessor sibling (immediate or
    # otherwise), then return that. If there are no siblings, traverse
    # parents until we get a valid path, and return that.

    if valid_path(model, path):
        return path

    new_path = tree_path_prev(path)
    while new_path:
        if valid_path(model, new_path):
            return new_path
        new_path = tree_path_prev(new_path)

    new_path = tree_path_up(path)
    while new_path:
        if valid_path(model, new_path):
            return new_path
        new_path = tree_path_up(new_path)


class SearchableTreeStore(Gtk.TreeStore):

    def inorder_search_down(self, it):
        while it:
            child = self.iter_children(it)
            if child:
                it = child
            else:
                next_it = self.iter_next(it)
                if next_it:
                    it = next_it
                else:
                    while True:
                        it = self.iter_parent(it)
                        if not it:
                            return
                        next_it = self.iter_next(it)
                        if next_it:
                            it = next_it
                            break
            yield it

    def inorder_search_up(self, it):
        while it:
            path = self.get_path(it)
            if path[-1]:
                path = path[:-1] + [path[-1] - 1]
                it = self.get_iter(path)
                while 1:
                    nc = self.iter_n_children(it)
                    if nc:
                        it = self.iter_nth_child(it, nc - 1)
                    else:
                        break
            else:
                up = self.iter_parent(it)
                if up:
                    it = up
                else:
                    return
            yield it

    def get_previous_next_paths(self, path, match_func):
        prev_path, next_path = None, None
        try:
            start_iter = self.get_iter(path)
        except ValueError:
            # Invalid tree path
            return None, None

        for it in self.inorder_search_up(start_iter):
            if match_func(it):
                prev_path = self.get_path(it)
                break

        for it in self.inorder_search_down(start_iter):
            if match_func(it):
                next_path = self.get_path(it)
                break

        return prev_path, next_path




import enum

from gi.repository import GtkSource

from meld.conf import _


class ActionMode(enum.IntEnum):
    """Action mode for chunk change actions"""
    Replace = 0
    Delete = 1
    Insert = 2


class ChunkAction(enum.Enum):

    delete = 'delete'
    replace = 'replace'
    copy_down = 'copy_down'
    copy_up = 'copy_up'


class FileComparisonMode(enum.Enum):
    AutoMerge = 'AutoMerge'
    Compare = 'Compare'


NEWLINES = {
    GtkSource.NewlineType.LF: ('\n', _("UNIX (LF)")),
    GtkSource.NewlineType.CR_LF: ('\r\n', _("DOS/Windows (CR-LF)")),
    GtkSource.NewlineType.CR: ('\r', _("Mac OS (CR)")),
}

FILE_FILTER_ACTION_FORMAT = 'folder-custom-filter-{}'
TEXT_FILTER_ACTION_FORMAT = 'text-custom-filter-{}'

#: Sentinel value for mtimes on files that don't exist.
MISSING_TIMESTAMP = -2147483648




from unittest import mock

import pytest

from meld.meldbuffer import BufferLines, MeldBuffer

text = ("""0
1
2
3
4
5
6
7
8
9
10""")


@pytest.fixture(scope='module', autouse=True)
def mock_bind_settings():
    with mock.patch('meld.meldbuffer.bind_settings', mock.DEFAULT):
        yield


@pytest.fixture
def buffer_setup():
    buf = MeldBuffer()
    buf.set_text(text)
    buffer_lines = BufferLines(buf)
    yield buf, buffer_lines


@pytest.mark.parametrize("line_start, line_end, expected_text", [
    (0, 1, ["0"],),
    (0, 2, ["0", "1"],),
    # zero-sized slice
    (9, 9, [],),
    (9, 10, ["9"],),
    (9, 11, ["9", "10"],),
    # Past the end of the buffer
    (9, 12, ["9", "10"],),
    # Waaaay past the end of the buffer
    (9, 9999, ["9", "10"],),
    # And sidling towards past-the-end start indices
    (10, 12, ["10"],),
    (11, 12, [],),
])
def test_meld_buffer_slicing(
        line_start, line_end, expected_text, buffer_setup):

    buffer, buffer_lines = buffer_setup
    assert buffer_lines[line_start:line_end] == expected_text


def test_meld_buffer_index_out_of_range(buffer_setup):

    buffer, buffer_lines = buffer_setup
    with pytest.raises(IndexError):
        buffer_lines[11]


def test_meld_buffer_cached_contents(buffer_setup):

    buffer, buffer_lines = buffer_setup
    text_lines = text.splitlines()
    assert len(buffer_lines.lines) == len(buffer_lines) == len(text_lines)

    # Check that without access, we have no cached contents
    assert buffer_lines.lines == [None] * len(text_lines)

    # Access the lines so that they're cached
    buffer_lines[:]

    # Note that this only happens to be true for our simple text; if
    # it were true in general, we wouldn't need the complexities of the
    # BufferLines class.
    assert buffer_lines.lines == text_lines


def test_meld_buffer_insert_text(buffer_setup):

    buffer, buffer_lines = buffer_setup

    # Access the lines so that they're cached
    buffer_lines[:]

    assert buffer_lines.lines[4:8] == ["4", "5", "6", "7"]

    # Delete from the start of line 5 to the start of line 7,
    # invalidating line 7 but leaving its contents intact.
    buffer.insert(
        buffer.get_iter_at_line(5),
        "hey\nthings",
    )
    assert buffer_lines.lines[4:8] == ["4", None, None, "6"]

    assert buffer_lines[5:7] == ["hey", "things5"]
    assert buffer_lines.lines[4:8] == ["4", "hey", "things5", "6"]


def test_meld_buffer_delete_range(buffer_setup):

    buffer, buffer_lines = buffer_setup

    # Access the lines so that they're cached
    buffer_lines[:]

    assert buffer_lines.lines[4:8] == ["4", "5", "6", "7"]

    # Delete from the start of line 5 to the start of line 7,
    # invalidating line 7 but leaving its contents intact.
    buffer.delete(
        buffer.get_iter_at_line(5),
        buffer.get_iter_at_line(7),
    )
    assert buffer_lines.lines[4:7] == ["4", None, "8"]

    assert buffer_lines[5] == "7"
    assert buffer_lines.lines[4:7] == ["4", "7", "8"]


def test_meld_buffer_cache_debug(caplog, buffer_setup):

    buffer, buffer_lines = buffer_setup
    buffer_lines = BufferLines(buffer, cache_debug=True)

    # Invalidate our line cache...
    buffer_lines.lines.append("invalid")

    # ...and check that insertion/deletion logs an error
    buffer.insert(
        buffer.get_iter_at_line(5),
        "hey",
    )
    assert len(caplog.records) == 1
    assert caplog.records[0].msg.startswith("Cache line count does not match")
    caplog.clear()

    buffer.delete(
        buffer.get_iter_at_line(5),
        buffer.get_iter_at_line(7),
    )
    assert len(caplog.records) == 1
    assert caplog.records[0].msg.startswith("Cache line count does not match")



# Copyright (C) 2014 Marco Brito <bcaza@null.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GObject, Gtk


class DiffGrid(Gtk.Grid):
    __gtype_name__ = "DiffGrid"

    column_count = 10
    handle_columns = (2, 6)

    def __init__(self):
        super().__init__()
        self._in_drag = False
        self._drag_pos = -1
        self._drag_handle = None
        self._handle1 = HandleWindow()
        self._handle2 = HandleWindow()

    def do_realize(self):
        Gtk.Grid.do_realize(self)
        self._handle1.realize(self)
        self._handle2.realize(self)

    def do_unrealize(self):
        self._handle1.unrealize()
        self._handle2.unrealize()
        Gtk.Grid.do_unrealize(self)

    def do_map(self):
        Gtk.Grid.do_map(self)

        drag = self.get_child_at(2, 0)
        self._handle1.set_visible(drag and drag.get_visible())
        drag = self.get_child_at(6, 0)
        self._handle2.set_visible(drag and drag.get_visible())

    def do_unmap(self):
        self._handle1.set_visible(False)
        self._handle2.set_visible(False)
        Gtk.Grid.do_unmap(self)

    def _handle_set_prelight(self, window, flag):
        if hasattr(window, "handle"):
            window.handle.set_prelight(flag)

    def do_enter_notify_event(self, event):
        if hasattr(event.window, "handle"):
            event.window.handle.set_prelight(True)

    def do_leave_notify_event(self, event):
        if self._in_drag:
            return

        if hasattr(event.window, "handle"):
            event.window.handle.set_prelight(False)

    def do_button_press_event(self, event):
        if event.button & Gdk.BUTTON_PRIMARY:
            self._drag_pos = event.x
            self._in_drag = True
            return True
        return False

    def do_button_release_event(self, event):
        if event.button & Gdk.BUTTON_PRIMARY:
            self._in_drag = False
            return True
        return False

    def do_motion_notify_event(self, event):
        if event.state & Gdk.ModifierType.BUTTON1_MASK:
            if hasattr(event.window, "handle"):
                x, y = event.window.get_position()
                pos = round(x + event.x - self._drag_pos)
                event.window.handle.set_position(pos)
                self._drag_handle = event.window.handle
                self.queue_resize_no_redraw()
                return True
        return False

    def _calculate_positions(
            self, xmin, xmax, pane_sep_width_1, pane_sep_width_2,
            wpane1, wpane2, wpane3):
        wremain = max(0, xmax - xmin - pane_sep_width_1 - pane_sep_width_2)
        pos1 = self._handle1.get_position(wremain, xmin)
        pos2 = self._handle2.get_position(wremain, xmin + pane_sep_width_1)

        if not self._drag_handle:
            npanes = 0
            if wpane1 > 0:
                npanes += 1
            if wpane2 > 0:
                npanes += 1
            if wpane3 > 0:
                npanes += 1
            wpane = float(wremain) / max(1, npanes)
            if wpane1 > 0:
                wpane1 = wpane
            if wpane2 > 0:
                wpane2 = wpane
            if wpane3 > 0:
                wpane3 = wpane

        xminlink1 = xmin + wpane1
        xmaxlink2 = xmax - wpane3 - pane_sep_width_2
        wlinkpane = pane_sep_width_1 + wpane2

        if wpane1 == 0:
            pos1 = xminlink1
        if wpane3 == 0:
            pos2 = xmaxlink2
        if wpane2 == 0:
            if wpane3 == 0:
                pos1 = pos2 - pane_sep_width_2
            else:
                pos2 = pos1 + pane_sep_width_1

        if self._drag_handle == self._handle2:
            xminlink2 = xminlink1 + wlinkpane
            pos2 = min(max(xminlink2, pos2), xmaxlink2)
            xmaxlink1 = pos2 - wlinkpane
            pos1 = min(max(xminlink1, pos1), xmaxlink1)
        else:
            xmaxlink1 = xmaxlink2 - wlinkpane
            pos1 = min(max(xminlink1, pos1), xmaxlink1)
            xminlink2 = pos1 + wlinkpane
            pos2 = min(max(xminlink2, pos2), xmaxlink2)

        self._handle1.set_position(pos1)
        self._handle2.set_position(pos2)
        return int(round(pos1)), int(round(pos2))

    def do_size_allocate(self, allocation):
        # We should be chaining up here to:
        #     Gtk.Grid.do_size_allocate(self, allocation)
        # However, when we do this, we hit issues with doing multiple
        # allocations in a single allocation cycle (see bgo#779883).

        self.set_allocation(allocation)
        wcols, hrows = self._get_min_sizes()
        yrows = [allocation.y,
                 allocation.y + hrows[0],
                 # Roughly equivalent to hard-coding row 1 to expand=True
                 allocation.y + (allocation.height - hrows[2] - hrows[3]),
                 allocation.y + (allocation.height - hrows[3]),
                 allocation.y + allocation.height]

        (wpane1, wgutter1, wlink1, wgutter2, wpane2, wgutter3, wlink2,
            wgutter4, wpane3, wmap) = wcols
        xmin = allocation.x
        xmax = allocation.x + allocation.width - wmap
        pane_sep_width_1 = wgutter1 + wlink1 + wgutter2
        pane_sep_width_2 = wgutter3 + wlink2 + wgutter4
        pos1, pos2 = self._calculate_positions(
            xmin, xmax, pane_sep_width_1, pane_sep_width_2,
            wpane1, wpane2, wpane3
        )
        wpane1 = pos1 - allocation.x
        wpane2 = pos2 - (pos1 + pane_sep_width_1)
        wpane3 = xmax - (pos2 + pane_sep_width_2)
        wcols = (
            allocation.x, wpane1, wgutter1, wlink1, wgutter2, wpane2,
            wgutter3, wlink2, wgutter4, wpane3, wmap)
        columns = [sum(wcols[:i + 1]) for i in range(len(wcols))]

        def child_allocate(child):
            if not child.get_visible():
                return
            left, top, width, height = self.child_get(
                child, 'left-attach', 'top-attach', 'width', 'height')
            # This is a copy, and we have to do this because there's no Python
            # access to Gtk.Allocation.
            child_alloc = self.get_allocation()
            child_alloc.x = columns[left]
            child_alloc.y = yrows[top]
            child_alloc.width = columns[left + width] - columns[left]
            child_alloc.height = yrows[top + height] - yrows[top]

            if self.get_direction() == Gtk.TextDirection.RTL:
                child_alloc.x = (
                    allocation.x + allocation.width -
                    (child_alloc.x - allocation.x) - child_alloc.width)

            child.size_allocate(child_alloc)

        for child in self.get_children():
            child_allocate(child)

        if self.get_realized():
            mapped = self.get_mapped()
            ydrag = yrows[0]
            hdrag = yrows[1] - yrows[0]
            self._handle1.set_visible(mapped and pane_sep_width_1 > 0)
            self._handle1.move_resize(pos1, ydrag, pane_sep_width_1, hdrag)
            self._handle2.set_visible(mapped and pane_sep_width_2 > 0)
            self._handle2.move_resize(pos2, ydrag, pane_sep_width_2, hdrag)

    def _get_min_sizes(self):
        hrows = [0] * 4
        wcols = [0] * self.column_count
        for row in range(4):
            for col in range(self.column_count):
                child = self.get_child_at(col, row)
                if child and child.get_visible():
                    msize, nsize = child.get_preferred_size()
                    # Ignore spanning columns in width calculations; we should
                    # do this properly, but it's difficult.
                    spanning = GObject.Value(int)
                    self.child_get_property(child, 'width', spanning)
                    spanning = spanning.get_int()
                    # We ignore natural size when calculating required
                    # width, but use it when doing required height. The
                    # logic here is that height-for-width means that
                    # minimum width requisitions mean more-than-minimum
                    # heights. This is all extremely dodgy, but works
                    # for now.
                    if spanning == 1:
                        wcols[col] = max(wcols[col], msize.width)
                    hrows[row] = max(hrows[row], msize.height, nsize.height)
        return wcols, hrows

    def do_draw(self, context):
        Gtk.Grid.do_draw(self, context)
        self._handle1.draw(context)
        self._handle2.draw(context)


class HandleWindow():

    # We restrict the handle width because render_handle doesn't pay
    # attention to orientation.
    handle_width = 10

    def __init__(self):
        self._widget = None
        self._window = None
        self._area_x = -1
        self._area_y = -1
        self._area_width = 1
        self._area_height = 1
        self._prelit = False
        self._pos = 0.0
        self._transform = (0, 0)

    def get_position(self, width, xtrans):
        self._transform = (width, xtrans)
        return float(self._pos * width) + xtrans

    def set_position(self, pos):
        width, xtrans = self._transform
        self._pos = float(pos - xtrans) / width

    def realize(self, widget):
        attr = Gdk.WindowAttr()
        attr.window_type = Gdk.WindowType.CHILD
        attr.x = self._area_x
        attr.y = self._area_y
        attr.width = self._area_width
        attr.height = self._area_height
        attr.wclass = Gdk.WindowWindowClass.INPUT_OUTPUT
        attr.event_mask = (widget.get_events() |
                           Gdk.EventMask.BUTTON_PRESS_MASK |
                           Gdk.EventMask.BUTTON_RELEASE_MASK |
                           Gdk.EventMask.ENTER_NOTIFY_MASK |
                           Gdk.EventMask.LEAVE_NOTIFY_MASK |
                           Gdk.EventMask.POINTER_MOTION_MASK)
        attr.cursor = Gdk.Cursor.new_for_display(widget.get_display(),
                                                 Gdk.CursorType.
                                                 SB_H_DOUBLE_ARROW)
        attr_mask = (Gdk.WindowAttributesType.X |
                     Gdk.WindowAttributesType.Y |
                     Gdk.WindowAttributesType.CURSOR)

        parent = widget.get_parent_window()
        self._window = Gdk.Window(parent, attr, attr_mask)
        self._window.handle = self
        self._widget = widget
        self._widget.register_window(self._window)

    def unrealize(self):
        self._widget.unregister_window(self._window)

    def set_visible(self, visible):
        if visible:
            self._window.show()
        else:
            self._window.hide()

    def move_resize(self, x, y, width, height):
        self._window.move_resize(x, y, width, height)
        self._area_x = x
        self._area_y = y
        self._area_width = width
        self._area_height = height

    def set_prelight(self, flag):
        self._prelit = flag
        self._widget.queue_draw_area(self._area_x, self._area_y,
                                     self._area_width, self._area_height)

    def draw(self, cairocontext):
        alloc = self._widget.get_allocation()
        padding = 5
        x = self._area_x - alloc.x + padding
        y = self._area_y - alloc.y + padding
        width = max(0, self._area_width - 2 * padding)
        height = max(0, self._area_height - 2 * padding)

        if width == 0 or height == 0:
            return

        stylecontext = self._widget.get_style_context()
        state = stylecontext.get_state()
        if self._widget.is_focus():
            state |= Gtk.StateFlags.SELECTED
        if self._prelit:
            state |= Gtk.StateFlags.PRELIGHT

        if Gtk.cairo_should_draw_window(cairocontext, self._window):
            stylecontext.save()
            stylecontext.set_state(state)
            stylecontext.add_class(Gtk.STYLE_CLASS_PANE_SEPARATOR)
            stylecontext.add_class(Gtk.STYLE_CLASS_VERTICAL)
            color = stylecontext.get_background_color(state)
            if color.alpha > 0.0:
                xcenter = x + width / 2.0 - self.handle_width / 2.0
                Gtk.render_handle(
                    stylecontext, cairocontext,
                    xcenter, y, self.handle_width, height)
            else:
                xcenter = x + width / 2.0
                Gtk.render_line(stylecontext, cairocontext,
                                xcenter, y, xcenter, y + height)
            stylecontext.restore()



# Copyright (C) 2011-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum

from gi.repository import Gio, GLib, GObject, Gtk

from meld.conf import _
from meld.melddoc import LabeledObjectMixin, MeldDoc
from meld.recent import recent_comparisons
from meld.ui.util import map_widgets_into_lists


class DiffType(enum.IntEnum):
    # TODO: This should probably live in MeldWindow
    Unselected = -1
    File = 0
    Folder = 1
    Version = 2

    def supports_blank(self):
        return self in (self.File, self.Folder)


@Gtk.Template(resource_path='/org/gnome/meld/ui/new-diff-tab.ui')
class NewDiffTab(Gtk.Alignment, LabeledObjectMixin):

    __gtype_name__ = "NewDiffTab"

    __gsignals__ = {
        'diff-created': (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    close_signal = MeldDoc.close_signal
    label_changed_signal = LabeledObjectMixin.label_changed

    label_text = _("New comparison")

    button_compare = Gtk.Template.Child()
    button_new_blank = Gtk.Template.Child()
    button_type_dir = Gtk.Template.Child()
    button_type_file = Gtk.Template.Child()
    button_type_vc = Gtk.Template.Child()
    choosers_notebook = Gtk.Template.Child()
    dir_chooser0 = Gtk.Template.Child()
    dir_chooser1 = Gtk.Template.Child()
    dir_chooser2 = Gtk.Template.Child()
    dir_three_way_checkbutton = Gtk.Template.Child()
    file_chooser0 = Gtk.Template.Child()
    file_chooser1 = Gtk.Template.Child()
    file_chooser2 = Gtk.Template.Child()
    file_three_way_checkbutton = Gtk.Template.Child()
    vc_chooser0 = Gtk.Template.Child()

    def __init__(self, parentapp):
        super().__init__()
        map_widgets_into_lists(
            self,
            ["file_chooser", "dir_chooser", "vc_chooser"]
        )
        self.button_types = [
            self.button_type_file,
            self.button_type_dir,
            self.button_type_vc,
        ]
        self.diff_methods = {
            DiffType.File: parentapp.append_filediff,
            DiffType.Folder: parentapp.append_dirdiff,
            DiffType.Version: parentapp.append_vcview,
        }
        self.diff_type = DiffType.Unselected

        default_path = GLib.get_home_dir()
        for chooser in self.file_chooser:
            chooser.set_current_folder(default_path)

        self.show()

    @Gtk.Template.Callback()
    def on_button_type_toggled(self, button, *args):
        if not button.get_active():
            if not any([b.get_active() for b in self.button_types]):
                button.set_active(True)
            return

        for b in self.button_types:
            if b is not button:
                b.set_active(False)

        self.diff_type = DiffType(self.button_types.index(button))
        self.choosers_notebook.set_current_page(self.diff_type + 1)
        # FIXME: Add support for new blank for VcView
        self.button_new_blank.set_sensitive(
            self.diff_type.supports_blank())
        self.button_compare.set_sensitive(True)

    @Gtk.Template.Callback()
    def on_three_way_checkbutton_toggled(self, button, *args):
        if button is self.file_three_way_checkbutton:
            self.file_chooser2.set_sensitive(button.get_active())
        else:  # button is self.dir_three_way_checkbutton
            self.dir_chooser2.set_sensitive(button.get_active())

    @Gtk.Template.Callback()
    def on_file_set(self, filechooser, *args):
        gfile = filechooser.get_file()
        if not gfile:
            return

        parent = gfile.get_parent()
        if not parent:
            return

        if parent.query_file_type(
                Gio.FileQueryInfoFlags.NONE, None) == Gio.FileType.DIRECTORY:
            for chooser in self.file_chooser:
                if not chooser.get_file():
                    chooser.set_current_folder_file(parent)

        # TODO: We could do checks here to prevent errors: check to see if
        # we've got binary files; check for null file selections; sniff text
        # encodings; check file permissions.

    def _get_num_paths(self):
        if self.diff_type in (DiffType.File, DiffType.Folder):
            three_way_buttons = (
                self.file_three_way_checkbutton,
                self.dir_three_way_checkbutton,
            )
            three_way = three_way_buttons[self.diff_type].get_active()
            num_paths = 3 if three_way else 2
        else:  # DiffType.Version
            num_paths = 1
        return num_paths

    @Gtk.Template.Callback()
    def on_button_compare_clicked(self, *args):
        type_choosers = (self.file_chooser, self.dir_chooser, self.vc_chooser)
        choosers = type_choosers[self.diff_type][:self._get_num_paths()]
        compare_gfiles = [chooser.get_file() for chooser in choosers]

        compare_kwargs = {}

        tab = self.diff_methods[self.diff_type](
            compare_gfiles, **compare_kwargs)
        recent_comparisons.add(tab)
        self.emit('diff-created', tab)

    @Gtk.Template.Callback()
    def on_button_new_blank_clicked(self, *args):
        # TODO: This doesn't work the way I'd like for DirDiff and VCView.
        # It should do something similar to FileDiff; give a tab with empty
        # file entries and no comparison done.

        # File comparison wants None for its paths here. Folder mode
        # needs an actual directory.
        if self.diff_type == DiffType.File:
            gfiles = [None] * self._get_num_paths()
        else:
            gfiles = [Gio.File.new_for_path("")] * self._get_num_paths()
        tab = self.diff_methods[self.diff_type](gfiles)
        self.emit('diff-created', tab)

    def on_container_switch_in_event(self, *args):
        self.label_changed.emit(self.label_text, self.tooltip_text)

    def on_container_switch_out_event(self, *args):
        pass

    def on_delete_event(self, *args):
        self.close_signal.emit(0)
        return Gtk.ResponseType.OK



# This file is part of the Hotwire Shell user interface.
#
# Copyright (C) 2007,2008 Colin Walters <walters@verbum.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Additional modifications made for use in Meld and adaptations for
# newer GTK+.
# Copyright (C) 2013 Kai Willadsen <kai.willadsen@gmail.com>

from gi.repository import Gtk, Pango

from meld.conf import _


def layout_text_and_icon(icon_name, primary_text, secondary_text=None):
    image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
    image.set_alignment(0.5, 0.5)

    vbox = Gtk.VBox(homogeneous=False, spacing=6)

    primary_label = Gtk.Label(
        label="<b>{}</b>".format(primary_text),
        wrap=True,
        wrap_mode=Pango.WrapMode.WORD_CHAR,
        use_markup=True,
        xalign=0,
        can_focus=True,
        selectable=True,
    )
    vbox.pack_start(primary_label, True, True, 0)

    if secondary_text:
        secondary_label = Gtk.Label(
            "<small>{}</small>".format(secondary_text),
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            use_markup=True,
            xalign=0,
            can_focus=True,
            selectable=True,
        )
        vbox.pack_start(secondary_label, True, True, 0)

    hbox_content = Gtk.HBox(homogeneous=False, spacing=8)
    hbox_content.pack_start(image, False, False, 0)
    hbox_content.pack_start(vbox, True, True, 0)
    hbox_content.show_all()
    return hbox_content


class MsgAreaController(Gtk.HBox):
    __gtype_name__ = "MsgAreaController"

    def __init__(self):
        super().__init__()

        self.__msgarea = None
        self.__msgid = None

    def has_message(self):
        return self.__msgarea is not None

    def get_msg_id(self):
        return self.__msgid

    def set_msg_id(self, msgid):
        self.__msgid = msgid

    def clear(self):
        if self.__msgarea is not None:
            self.remove(self.__msgarea)
            self.__msgarea.destroy()
            self.__msgarea = None
        self.__msgid = None

    def new_from_text_and_icon(
            self, icon_name, primary, secondary=None, buttons=None):
        self.clear()
        msgarea = self.__msgarea = Gtk.InfoBar()

        if buttons:
            for (text, respid) in buttons:
                self.add_button(text, respid)

        content = layout_text_and_icon(icon_name, primary, secondary)

        content_area = msgarea.get_content_area()
        content_area.foreach(content_area.remove, None)
        content_area.add(content)

        self.pack_start(msgarea, True, True, 0)
        return msgarea

    def add_dismissable_msg(self, icon, primary, secondary, close_panes=None):
        def clear_all(*args):
            if close_panes:
                for pane in close_panes:
                    pane.clear()
            else:
                self.clear()
        msgarea = self.new_from_text_and_icon(icon, primary, secondary)
        msgarea.add_button(_("Hi_de"), Gtk.ResponseType.CLOSE)
        msgarea.connect("response", clear_all)
        msgarea.show_all()
        return msgarea

    def add_action_msg(self, icon, primary, secondary, action_label, callback):
        def on_response(msgarea, response_id, *args):
            self.clear()
            if response_id == Gtk.ResponseType.ACCEPT:
                callback()

        msgarea = self.new_from_text_and_icon(icon, primary, secondary)
        msgarea.add_button(action_label, Gtk.ResponseType.ACCEPT)
        msgarea.add_button(_("Hi_de"), Gtk.ResponseType.CLOSE)
        msgarea.connect("response", on_response)
        msgarea.show_all()
        return msgarea




# The original C implementation is part of the GTK+ project, under
#   gtk+/demos/gtk-demo/foreigndrawing.c
#
# This Python port of portions of the original source code is copyright
# (C) 2009-2015 Kai Willadsen <kai.willadsen@gmail.com>, and is released
# under the same LGPL version 2 (or later) license.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <http://www.gnu.org/licenses/>.


import logging
import re

from gi.repository import GObject, Gtk

log = logging.getLogger(__name__)


def append_element(path, selector):
    pseudo_classes = [
        ('active',        Gtk.StateFlags.ACTIVE),
        ('hover',         Gtk.StateFlags.PRELIGHT),
        ('selected',      Gtk.StateFlags.SELECTED),
        ('disabled',      Gtk.StateFlags.INSENSITIVE),
        ('indeterminate', Gtk.StateFlags.INCONSISTENT),
        ('focus',         Gtk.StateFlags.FOCUSED),
        ('backdrop',      Gtk.StateFlags.BACKDROP),
        ('dir(ltr)',      Gtk.StateFlags.DIR_LTR),
        ('dir(rtl)',      Gtk.StateFlags.DIR_RTL),
        ('link',          Gtk.StateFlags.LINK),
        ('visited',       Gtk.StateFlags.VISITED),
        ('checked',       Gtk.StateFlags.CHECKED),
        ('drop(active)',  Gtk.StateFlags.DROP_ACTIVE)
    ]

    toks = [t for t in re.split(r'([#\.:])', selector) if t]
    elements = [toks[i] + toks[i + 1] for i in range(1, len(toks), 2)]

    name = toks[0]
    if name[0].isupper():
        gtype = GObject.GType.from_name(name)
        if gtype == GObject.TYPE_INVALID:
            log.error('Unknown type name "%s"', name)
            return
        path.append_type(gtype)
    else:
        # Omit type, we're using name
        path.append_type(GObject.TYPE_NONE)
        path.iter_set_object_name(-1, name)

    for segment in elements:
        segment_type = segment[0]
        name = segment[1:]
        if segment_type == '#':
            path.iter_set_name(-1, name)
            break
        elif segment_type == '.':
            path.iter_add_class(-1, name)
            break
        elif segment_type == ':':
            for class_name, class_state in pseudo_classes:
                if name == class_name:
                    path.iter_set_state(
                        -1, path.iter_get_state(-1) | class_state)
                    break
            else:
                log.error('Unknown pseudo-class :%s', name)
                pass
            break
        else:
            assert False


def create_context_for_path(path, parent):
    context = Gtk.StyleContext.new()
    context.set_path(path)
    context.set_parent(parent)
    context.set_state(path.iter_get_state(-1))
    return context


def get_style(parent, selector):
    if parent:
        path = Gtk.WidgetPath.copy(parent.get_path())
    else:
        path = Gtk.WidgetPath.new()
    append_element(path, selector)
    return create_context_for_path(path, parent)


def query_size(context, width, height):
    margin = context.get_margin(context.get_state())
    border = context.get_border(context.get_state())
    padding = context.get_padding(context.get_state())
    min_width = context.get_property('min-width', context.get_state())
    min_height = context.get_property('min-height', context.get_state())
    min_width += (
        margin.left + margin.right + border.left + border.right +
        padding.left + padding.right)
    min_height += (
        margin.top + margin.bottom + border.top + border.bottom +
        padding.top + padding.bottom)
    return max(width, min_width), max(height, min_height)


def draw_style_common(context, cr, x, y, width, height):

    margin = context.get_margin(context.get_state())
    border = context.get_border(context.get_state())
    padding = context.get_padding(context.get_state())

    min_width = context.get_property('min-width', context.get_state())
    min_height = context.get_property('min-height', context.get_state())

    x += margin.left
    y += margin.top
    width -= margin.left + margin.right
    height -= margin.top + margin.bottom

    width = max(width, min_width)
    height = max(height, min_height)

    Gtk.render_background(context, cr, x, y, width, height)
    Gtk.render_frame(context, cr, x, y, width, height)

    contents_x = x + border.left + padding.left
    contents_y = y + border.top + padding.top
    contents_width = (
        width - border.left - border.right - padding.left - padding.right)
    contents_height = (
        height - border.top - border.bottom - padding.top - padding.bottom)

    return contents_x, contents_y, contents_width, contents_height



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2010-2011 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module to help implement undo functionality.

Usage:

t = TextWidget()
s = UndoSequence()
def on_textwidget_text_inserted():
    s.begin_group()
    if not t.is_modified():
        s.add_action( TextWidgetModifiedAction() )
    s.add_action( InsertionAction() )
    s.end_group()

def on_undo_button_pressed():
    s.undo()
"""

import logging
import weakref

from gi.repository import GObject

log = logging.getLogger(__name__)


class GroupAction:
    """A group action combines several actions into one logical action.
    """
    def __init__(self, seq):
        self.seq = seq
        # TODO: If a GroupAction affects more than one sequence, our logic
        # breaks. Currently, this isn't a problem.
        self.buffer = seq.actions[0].buffer

    def undo(self):
        actions = []
        while self.seq.can_undo():
            actions.extend(self.seq.undo())
        return actions

    def redo(self):
        actions = []
        while self.seq.can_redo():
            actions.extend(self.seq.redo())
        return actions


class UndoSequence(GObject.GObject):
    """A manager class for operations which can be undone/redone.
    """

    __gsignals__ = {
        'can-undo': (
            GObject.SignalFlags.RUN_FIRST,
            None, (GObject.TYPE_BOOLEAN,)
        ),
        'can-redo': (
            GObject.SignalFlags.RUN_FIRST,
            None, (GObject.TYPE_BOOLEAN,)
        ),
        'checkpointed': (
            GObject.SignalFlags.RUN_FIRST,
            None, (GObject.TYPE_OBJECT, GObject.TYPE_BOOLEAN,)
        ),
    }

    def __init__(self, buffers):
        """Create an empty UndoSequence

        An undo sequence is tied to a collection of GtkTextBuffers, and
        expects to maintain undo checkpoints for the same set of
        buffers for the lifetime of the UndoSequence.
        """
        super().__init__()
        self.buffer_refs = [weakref.ref(buf) for buf in buffers]
        self.clear()

    def clear(self):
        """Remove all undo and redo actions from this sequence

        If the sequence was previously able to undo and/or redo, the
        'can-undo' and 'can-redo' signals are emitted.
        """
        if self.can_undo():
            self.emit('can-undo', 0)
        if self.can_redo():
            self.emit('can-redo', 0)
        self.actions = []
        self.next_redo = 0
        self.checkpoints = {
            # Each buffer's checkpoint starts at zero and has no end
            ref(): [0, None] for ref in self.buffer_refs
        }
        self.group = None
        self.busy = False

    def can_undo(self):
        """Return whether an undo is possible."""
        return getattr(self, 'next_redo', 0) > 0

    def can_redo(self):
        """Return whether a redo is possible."""
        next_redo = getattr(self, 'next_redo', 0)
        return next_redo < len(getattr(self, 'actions', []))

    def add_action(self, action):
        """Add an action to the undo list.

        Arguments:

        action -- A class with two callable attributes: 'undo' and 'redo'
                  which are called by this sequence during an undo or redo.
        """
        if self.busy:
            return

        if self.group is None:
            if self.checkpointed(action.buffer):
                self.checkpoints[action.buffer][1] = self.next_redo
                self.emit('checkpointed', action.buffer, False)
            else:
                # If we go back in the undo stack before the checkpoint starts,
                # and then modify the buffer, we lose the checkpoint altogether
                start, end = self.checkpoints.get(action.buffer, (None, None))
                if start is not None and start > self.next_redo:
                    self.checkpoints[action.buffer] = (None, None)
            could_undo = self.can_undo()
            could_redo = self.can_redo()
            self.actions[self.next_redo:] = []
            self.actions.append(action)
            self.next_redo += 1
            if not could_undo:
                self.emit('can-undo', 1)
            if could_redo:
                self.emit('can-redo', 0)
        else:
            self.group.add_action(action)

    def undo(self):
        """Undo an action.

        Raises an AssertionError if the sequence is not undoable.
        """
        assert self.next_redo > 0
        self.busy = True
        buf = self.actions[self.next_redo - 1].buffer
        if self.checkpointed(buf):
            self.emit('checkpointed', buf, False)
        could_redo = self.can_redo()
        self.next_redo -= 1
        actions = self.actions[self.next_redo].undo()
        self.busy = False
        if not self.can_undo():
            self.emit('can-undo', 0)
        if not could_redo:
            self.emit('can-redo', 1)
        if self.checkpointed(buf):
            self.emit('checkpointed', buf, True)
        return actions

    def redo(self):
        """Redo an action.

        Raises and AssertionError if the sequence is not undoable.
        """
        assert self.next_redo < len(self.actions)
        self.busy = True
        buf = self.actions[self.next_redo].buffer
        if self.checkpointed(buf):
            self.emit('checkpointed', buf, False)
        could_undo = self.can_undo()
        a = self.actions[self.next_redo]
        self.next_redo += 1
        actions = a.redo()
        self.busy = False
        if not could_undo:
            self.emit('can-undo', 1)
        if not self.can_redo():
            self.emit('can-redo', 0)
        if self.checkpointed(buf):
            self.emit('checkpointed', buf, True)
        return actions

    def checkpoint(self, buf):
        start = self.next_redo
        while start > 0 and self.actions[start - 1].buffer != buf:
            start -= 1
        end = self.next_redo
        while (end < len(self.actions) - 1 and
               self.actions[end + 1].buffer != buf):
            end += 1
        if end == len(self.actions):
            end = None
        self.checkpoints[buf] = [start, end]
        self.emit('checkpointed', buf, True)

    def checkpointed(self, buf):
        # While the main undo sequence should always have checkpoints
        # recorded, grouped subsequences won't.
        start, end = self.checkpoints.get(buf, (None, None))
        if start is None:
            return False
        if end is None:
            end = len(self.actions)
        return start <= self.next_redo <= end

    def begin_group(self):
        """Group several actions into a single logical action.

        When you wrap several calls to add_action() inside begin_group()
        and end_group(), all the intervening actions are considered
        one logical action. For instance a 'replace' action may be
        implemented as a pair of 'delete' and 'create' actions, but
        undoing should undo both of them.
        """
        if self.busy:
            return

        if self.group:
            self.group.begin_group()
        else:
            buffers = [ref() for ref in self.buffer_refs]
            self.group = UndoSequence(buffers)

    def end_group(self):
        """End a logical group action

        This must always be paired with a begin_group() call. However,
        we don't complain if this is not the case because we rely on
        external libraries (i.e., GTK+ and GtkSourceView) also pairing
        these correctly.

        See also begin_group().
        """
        if self.busy:
            return

        if self.group is None:
            log.warning('Tried to end a non-existent group')
            return

        if self.group.group is not None:
            self.group.end_group()
        else:
            group = self.group
            self.group = None
            # Collapse single action groups
            if len(group.actions) == 1:
                self.add_action(group.actions[0])
            elif len(group.actions) > 1:
                self.add_action(GroupAction(group))

    def abort_group(self):
        """Clear the currently grouped actions

        This discards all actions since the last begin_group() was
        called. Note that it does not actually undo the actions
        themselves.
        """
        if self.busy:
            return

        if self.group is None:
            log.warning('Tried to abort a non-existent group')
            return

        if self.group.group is not None:
            self.group.abort_group()
        else:
            self.group = None

    def in_grouped_action(self):
        return self.group is not None




from unittest import mock

import pytest

from meld.misc import all_same, calc_syncpoint, merge_intervals


@pytest.mark.parametrize("intervals, expected", [
    # Dominated by a single range
    ([(1, 5), (5, 9), (10, 11), (0, 20)], [(0, 20)]),
    # No overlap
    ([(1, 5), (6, 9), (10, 11)], [(1, 5), (6, 9), (10, 11)]),
    # Two overlap points between ranges
    ([(1, 5), (5, 9), (10, 12), (11, 20)], [(1, 9), (10, 20)]),
    # Two overlap points between ranges, out of order
    ([(5, 9), (1, 5), (11, 20), (10, 12)], [(1, 9), (10, 20)]),
    # Two equal ranges
    ([(1, 5), (7, 8), (1, 5)], [(1, 5), (7, 8)]),
    # Three ranges overlap
    ([(1, 5), (4, 10), (9, 15)], [(1, 15)])
])
def test_merge_intervals(intervals, expected):
    merged = merge_intervals(intervals)
    assert merged == expected


@pytest.mark.parametrize("value, page_size, lower, upper, expected", [
    # Boring top
    (0, 100, 0, 1000, 0.0),
    # Above the top!
    (0, 100, 100, 1000, 0.0),
    # Normal top scaling
    (25, 100, 0, 1000, 0.25),
    (50, 100, 0, 1000, 0.5),
    # Scaling with a lower offset
    (25, 100, 25, 1000, 0.0),
    (50, 100, 25, 1000, 0.25),
    # Somewhere in the middle
    (500, 100, 0, 1000, 0.5),
    # Normal bottom scaling
    (850, 100, 0, 1000, 0.5),
    (875, 100, 0, 1000, 0.75),
    # Boring bottom
    (900, 100, 0, 1000, 1.0),
    # Below the bottom!
    (1100, 100, 0, 1000, 1.0),
])
def test_calc_syncpoint(value, page_size, lower, upper, expected):
    import gi
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk
    adjustment = Gtk.Adjustment()
    adjustment.configure(value, lower, upper, 1, 1, page_size)
    syncpoint = calc_syncpoint(adjustment)
    assert syncpoint == expected


@pytest.mark.parametrize("lst, expected", [
    (None, True),
    ([], True),
    ([0], True),
    ([1], True),
    ([0, 0], True),
    ([0, 1], False),
    ([1, 0], False),
    ([1, 1], True),
    ([0, 0, 0], True),
    ([0, 0, 1], False),
    ([0, 1, 0], False),
    ([0, 1, 1], False),
    ([1, 0, 0], False),
    ([1, 0, 1], False),
    ([1, 1, 0], False),
    ([1, 1, 1], True)
])
def test_all_same(lst, expected):
    assert all_same(lst) == expected


@pytest.mark.parametrize("os_name, paths, expected", [
    ('posix', ['/tmp/foo1', '/tmp/foo2'], ['foo1', 'foo2']),
    ('posix', ['/tmp/foo1', '/tmp/foo2', '/tmp/foo3'], ['foo1', 'foo2', 'foo3']),
    ('posix', ['/tmp/bar/foo1', '/tmp/woo/foo2'], ['foo1', 'foo2']),
    ('posix', ['/tmp/bar/foo1', '/tmp/woo/foo1'], ['[bar] foo1', '[woo] foo1']),
    ('posix', ['/tmp/bar/foo1', '/tmp/woo/foo1', '/tmp/ree/foo1'], ['[bar] foo1', '[woo] foo1', '[ree] foo1']),
    ('posix', ['/tmp/bar/deep/deep', '/tmp/bar/shallow'], ['deep', 'shallow']),
    ('posix', ['/tmp/bar/deep/deep/foo1', '/tmp/bar/shallow/foo1'], ['[deep] foo1', '[shallow] foo1']),
    # This case doesn't actually make much sense, so it's not that bad
    # that our output is... somewhat unclear.
    ('posix', ['/tmp/bar/subdir/subsub', '/tmp/bar/'], ['subsub', 'bar']),
    ('nt', ['C:\\Users\\hmm\\bar', 'C:\\Users\\hmm\\foo'], ['bar', 'foo']),
    ('nt', ['C:\\Users\\bar\\hmm', 'C:\\Users\\foo\\hmm'], ['[bar] hmm', '[foo] hmm']),
    # Check that paths with no commonality are handled
    ('posix', ['nothing in', 'common'], ['nothing in', 'common']),
    ('posix', ['<unnamed>', '/tmp/real/path'], ['<unnamed>', '/tmp/real/path']),
])
def test_shorten_names(os_name, paths, expected):
    from meld.misc import shorten_names

    with mock.patch('os.name', os_name):
        assert shorten_names(*paths) == expected



# Copyright (C) 2013-2014 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math

from gi.repository import Gdk, GtkSource, Pango

from meld.settings import get_meld_settings
from meld.style import get_common_theme


def get_background_rgba(renderer):
    '''Get and cache the expected background for the renderer widget

    Current versions of GTK+ don't paint the background of text view
    gutters with the actual expected widget background, which causes
    them to look wrong when put next to any other widgets. This hack
    just gets the background from the renderer's view, and then caches
    it for performance, and on the basis that all renderers will be
    assigned to similarly-styled views. This is fragile, but the
    alternative is really significantly slower.
    '''
    global _background_rgba
    if _background_rgba is None:
        if renderer.props.view:
            stylecontext = renderer.props.view.get_style_context()
            background_set, _background_rgba = (
                stylecontext.lookup_color('theme_bg_color'))
    return _background_rgba


_background_rgba = None


class MeldGutterRenderer:

    def set_renderer_defaults(self):
        self.set_alignment_mode(GtkSource.GutterRendererAlignmentMode.FIRST)
        self.set_padding(3, 0)
        self.set_alignment(0.5, 0.5)

    def on_setting_changed(self, settings, key):
        if key == 'style-scheme':
            self.fill_colors, self.line_colors = get_common_theme()
            alpha = self.fill_colors['current-chunk-highlight'].alpha
            self.chunk_highlights = {
                state: Gdk.RGBA(*[alpha + c * (1.0 - alpha) for c in colour])
                for state, colour in self.fill_colors.items()
            }

    def draw_chunks(
            self, context, background_area, cell_area, start, end, state):

        chunk = self._chunk
        if not chunk:
            return

        line = start.get_line()
        is_first_line = line == chunk[1]
        is_last_line = line == chunk[2] - 1
        if not (is_first_line or is_last_line):
            # Only paint for the first and last lines of a chunk
            return

        x = background_area.x - 1
        y = background_area.y
        width = background_area.width + 2
        height = 1 if chunk[1] == chunk[2] else background_area.height

        context.set_line_width(1.0)
        Gdk.cairo_set_source_rgba(context, self.line_colors[chunk[0]])
        if is_first_line:
            context.move_to(x, y + 0.5)
            context.rel_line_to(width, 0)
        if is_last_line:
            context.move_to(x, y - 0.5 + height)
            context.rel_line_to(width, 0)
        context.stroke()

    def query_chunks(self, start, end, state):
        line = start.get_line()
        chunk_index = self.linediffer.locate_chunk(self.from_pane, line)[0]
        in_chunk = chunk_index is not None

        chunk = None
        if in_chunk:
            chunk = self.linediffer.get_chunk(
                chunk_index, self.from_pane, self.to_pane)

        if chunk is not None:
            if chunk[1] == chunk[2]:
                background_rgba = get_background_rgba(self)
            elif self.props.view.current_chunk_check(chunk):
                background_rgba = self.chunk_highlights[chunk[0]]
            else:
                background_rgba = self.fill_colors[chunk[0]]
        else:
            # TODO: Remove when fixed in upstream GTK+
            background_rgba = get_background_rgba(self)
        self._chunk = chunk
        self.set_background(background_rgba)
        return in_chunk


# GutterRendererChunkLines is an adaptation of GtkSourceGutterRendererLines
# Copyright (C) 2010 - Jesse van den Kieboom
#
# Python reimplementation is Copyright (C) 2015 Kai Willadsen


class GutterRendererChunkLines(
        GtkSource.GutterRendererText, MeldGutterRenderer):
    __gtype_name__ = "GutterRendererChunkLines"

    def __init__(self, from_pane, to_pane, linediffer):
        super().__init__()
        self.set_renderer_defaults()
        self.from_pane = from_pane
        self.to_pane = to_pane
        # FIXME: Don't pass in the linediffer; pass a generator like elsewhere
        self.linediffer = linediffer

        self.num_line_digits = 0
        self.changed_handler_id = None

        meld_settings = get_meld_settings()
        meld_settings.connect('changed', self.on_setting_changed)
        self.on_setting_changed(meld_settings, 'style-scheme')

    def do_change_buffer(self, old_buffer):
        if old_buffer:
            old_buffer.disconnect(self.changed_handler_id)

        view = self.get_view()
        if view:
            buf = view.get_buffer()
            if buf:
                self.changed_handler_id = buf.connect(
                    "changed", self.recalculate_size)
                self.recalculate_size(buf)

    def _measure_markup(self, markup):
        layout = self.get_view().create_pango_layout()
        layout.set_markup(markup)
        w, h = layout.get_size()
        return w / Pango.SCALE, h / Pango.SCALE

    def recalculate_size(self, buf):

        # Always calculate display size for at least two-digit line counts
        num_lines = max(buf.get_line_count(), 99)
        num_digits = int(math.ceil(math.log(num_lines, 10)))

        if num_digits == self.num_line_digits:
            return

        self.num_line_digits = num_digits
        markup = "<b>%d</b>" % num_lines
        width, height = self._measure_markup(markup)
        self.set_size(width)

    def do_draw(self, context, background_area, cell_area, start, end, state):
        GtkSource.GutterRendererText.do_draw(
            self, context, background_area, cell_area, start, end, state)
        self.draw_chunks(
            context, background_area, cell_area, start, end, state)

    def do_query_data(self, start, end, state):
        self.query_chunks(start, end, state)
        line = start.get_line() + 1
        current_line = state & GtkSource.GutterRendererState.CURSOR
        markup = "<b>%d</b>" % line if current_line else str(line)
        self.set_markup(markup, -1)




import importlib.machinery
import importlib.util
import sys
from unittest import mock

import pytest


@pytest.fixture(autouse=True)
def default_icon_theme():
    # Our tests need to run on a system with no default display, so all
    # our display-specific get_default() stuff will break.

    from gi.repository import Gtk
    with mock.patch(
            'gi.repository.Gtk.IconTheme.get_default',
            mock.Mock(spec=Gtk.IconTheme.get_default)):
        yield


@pytest.fixture(autouse=True)
def template_resources():
    import gi  # noqa: F401
    with mock.patch(
            'gi._gtktemplate.validate_resource_path',
            mock.Mock(return_value=True)):
        yield


def import_meld_conf():
    loader = importlib.machinery.SourceFileLoader(
        'meld.conf', './meld/conf.py.in')
    spec = importlib.util.spec_from_loader(loader.name, loader)
    mod = importlib.util.module_from_spec(spec)
    loader.exec_module(mod)

    import meld
    meld.conf = mod
    sys.modules['meld.conf'] = mod


import_meld_conf()



#!/usr/bin/env python3

import glob
import pathlib
from distutils.core import setup

# Copy conf.py in place if necessary
base_path = pathlib.Path(__file__).parent
conf_path = base_path / 'meld' / 'conf.py'

if not conf_path.exists():
    import shutil
    shutil.copyfile(conf_path.with_suffix('.py.in'), conf_path)

import meld.build_helpers  # noqa:E402 isort:skip
import meld.conf  # noqa:E402 isort:skip

setup(
    name=meld.conf.__package__,
    version=meld.conf.__version__,
    description='Visual diff and merge tool',
    author='Kai Willadsen',
    author_email='kai.willadsen@gmail.com',
    url='http://meldmerge.org',
    license='GPLv2+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Desktop Environment :: Gnome',
        'Topic :: Software Development',
        'Topic :: Software Development :: Version Control',
    ],
    keywords=['diff', 'merge'],
    packages=[
        'meld',
        'meld.matchers',
        'meld.ui',
        'meld.vc',
    ],
    package_data={
        'meld': ['README', 'COPYING', 'NEWS'],
        'meld.vc': ['README', 'COPYING'],
    },
    scripts=['bin/meld'],
    data_files=[
        ('share/man/man1',
         ['data/meld.1']
         ),
        ('share/doc/meld-' + meld.conf.__version__,
         ['COPYING', 'NEWS']
         ),
        ('share/meld/icons',
         glob.glob("data/icons/*.png") +
         glob.glob("data/icons/COPYING*")
         ),
        ('share/meld/styles',
         glob.glob("data/styles/*.xml")
         ),
        ('share/meld/ui',
         glob.glob("data/ui/*.ui") + glob.glob("data/ui/*.xml")
         ),
    ],
    cmdclass={
        "build_i18n": meld.build_helpers.build_i18n,
        "build_help": meld.build_helpers.build_help,
        "build_icons": meld.build_helpers.build_icons,
        "build_data": meld.build_helpers.build_data,
        "build_py": meld.build_helpers.build_py,
        "install": meld.build_helpers.install,
        "install_data": meld.build_helpers.install_data,
    },
    distclass=meld.build_helpers.MeldDistribution,
)



from enum import Enum
from os import path

import pytest

from .fixture import make

DiffResult = Enum('DiffResult', 'Same SameFiltered DodgySame DodgyDifferent Different FileError')


@pytest.fixture
def differnt_dirs():
    make()


def abspath(*args):
    d = path.dirname(__file__)
    return list(path.join(d, arg) for arg in args)


cmp_args = {
    'shallow-comparison': False,
    'time-resolution': 10000000000,
    'ignore_blank_lines': True,
    'apply-text-filters': True
}

no_ignore_args = dict(cmp_args)
no_ignore_args['ignore_blank_lines'] = False
no_ignore_args['apply-text-filters'] = False

dodgy_args = dict(cmp_args)
dodgy_args['shallow-comparison'] = True


@pytest.mark.parametrize('files, regexes, comparison_args, expected', [
    # empty file list
    ((), [], cmp_args, DiffResult.Same),
    # dirs are same
    (('diffs/a', 'diffs/b'), [], cmp_args, DiffResult.Same),
    # dir and file ar diffent
    (('diffs/a', 'diffs/b/b.txt'), [], cmp_args, DiffResult.Different),
    # shallow equal (time + size)
    (('diffs/a/d/d.txt', 'diffs/b/d/d.1.txt'), [], dodgy_args, DiffResult.DodgySame),
    # empty files (fastest equal, wont read files)
    (('diffs/a/c/c.txt', 'diffs/b/c/c.txt'), [], cmp_args, DiffResult.Same),
    # 4.1kb vs 4.1kb file (slow equal, read both until end)
    (('diffs/a/d/d.txt', 'diffs/b/d/d.txt'), [], cmp_args, DiffResult.Same),
    # 4.1kb vs 4.1kb file (fast different, first chunk diff)
    (('diffs/a/d/d.txt', 'diffs/b/d/d.1.txt'), [], cmp_args, DiffResult.Different),
    # 4.1kb vs 4.1kb file (slow different, read both until end)
    (('diffs/a/d/d.txt', 'diffs/b/d/d.2.txt'), [], cmp_args, DiffResult.Different),
    # empty vs 1b file (fast different, first chunk diff)
    (('diffs/a/e/g/g.txt', 'diffs/b/e/g/g.txt'), [], cmp_args, DiffResult.Different),
    # CRLF vs CRLF with trailing, ignoring blank lines
    (('diffs/a/crlf.txt', 'diffs/a/crlftrailing.txt'), [], cmp_args, DiffResult.SameFiltered),
    # CRLF vs CRLF with trailing, not ignoring blank lines
    (('diffs/a/crlf.txt', 'diffs/a/crlftrailing.txt'), [], no_ignore_args, DiffResult.Different),
    # LF vs LF with trailing, ignoring blank lines
    (('diffs/b/lf.txt', 'diffs/b/lftrailing.txt'), [], cmp_args, DiffResult.SameFiltered),
    # LF vs LF with trailing, not ignoring blank lines
    (('diffs/b/lf.txt', 'diffs/b/lftrailing.txt'), [], no_ignore_args, DiffResult.Different),
    # CRLF vs LF, ignoring blank lines
    (('diffs/a/crlf.txt', 'diffs/b/lf.txt'), [], cmp_args, DiffResult.SameFiltered),
    # CRLF vs LF, not ignoring blank lines
    (('diffs/a/crlf.txt', 'diffs/b/lf.txt'), [], no_ignore_args, DiffResult.Different),
    # CRLF with trailing vs LF with trailing, ignoring blank lines
    (('diffs/a/crlftrailing.txt', 'diffs/b/lftrailing.txt'), [], cmp_args, DiffResult.SameFiltered),
    # CRLF with trailing vs LF with trailing, not ignoring blank lines
    (('diffs/a/crlftrailing.txt', 'diffs/b/lftrailing.txt'), [], no_ignore_args, DiffResult.Different),
])
def test_files_same(files, regexes, comparison_args, expected, differnt_dirs):
    from meld.dirdiff import _files_same

    files_path = abspath(*files)
    result = _files_same(files_path, regexes, comparison_args)
    actual = DiffResult(result + 1)
    assert actual == expected



#!/usr/bin/env python3

import pathlib
import sys


def main():
    in_path = pathlib.Path(sys.argv[1])
    out_path = pathlib.Path(sys.argv[2])
    if not in_path.exists():
        print(f"Couldn't find {in_path}")
        sys.exit(1)

    lines = in_path.read_text().splitlines(keepends=True)
    lines[0] = "".join(lines[0].split("env "))
    out_path.write_text("".join(lines))


if __name__ == "__main__":
    main()
